#include <sstream>
#include <algorithm>
#include <vector>
#include <future>
#include <limits>
#include <fstream>

#include "enigma/enigma.h"

#include "fitness/fitness.hpp"

using namespace sinkov;

constexpr size_t NUMBER_TO_HOLD = 3000;

#define MINFINITY -std::numeric_limits<double>::infinity()

std::array<std::string, 8> rotorNames = {"I", "II", "III", "IV", "V", "VI", "VII", "VIII"};

char int_to_char(int c)
{
    return static_cast<char>(c + 'A');
}

struct Setting
{
    std::array<int, 3> rotorIndexes{0, 0, 0};
    std::array<int, 3> rotorPositions{0, 0, 0};
    std::array<int, 3> ringSettings{0, 0, 0};
    std::vector<std::string> plugs{};
    std::string text;
    double score{MINFINITY};

    std::ostream &print(std::ostream &s, bool showScore = true) const
    {
        s << rotorNames[rotorIndexes[0]] << " " << rotorNames[rotorIndexes[1]] << " " << rotorNames[rotorIndexes[2]];
        s << " " << int_to_char(ringSettings[0]) << int_to_char(ringSettings[1]) << int_to_char(ringSettings[2]);
        for (const auto &plug : plugs)
        {
            s << " " << plug;
        }
        s << " " << int_to_char(rotorPositions[0]) << int_to_char(rotorPositions[1]) << int_to_char(rotorPositions[2]);
        // s << " - " << set.text;
        if (showScore)
        {
            s << " : " << score;
        }
        return s;
    }
};

std::ostream &operator<<(std::ostream &s, const Setting &set)
{
    return set.print(s);
}

void try_enigma(Setting &setting, const std::string &cipher, bool useQuadgrams = false)
{
    Rotor left_rotor = rotors.at(rotorNames[setting.rotorIndexes[0]]).Duplicate(setting.ringSettings[0], setting.rotorPositions[0]);
    Rotor center_rotor = rotors.at(rotorNames[setting.rotorIndexes[1]]).Duplicate(setting.ringSettings[1], setting.rotorPositions[1]);
    Rotor right_rotor = rotors.at(rotorNames[setting.rotorIndexes[2]]).Duplicate(setting.ringSettings[2], setting.rotorPositions[2]);

    Enigma enigma({left_rotor, center_rotor, right_rotor}, B, Plugboard(setting.plugs));

    setting.text = enigma.encrypt(cipher);
    if (!useQuadgrams)
        setting.score = calculate_fitness(setting.text);
    else
        setting.score = quadgrams::calculate_fitness(setting.text);
}

std::vector<Setting> try_all_keys(int firstRotor, int secondRotor, int thirdRotor, const std::string &cipher)
{
    std::vector<std::string> pairs;
    for (size_t i = 0; i < 10; i++)
    {
        pairs.push_back("AA");
    }

    Plugboard plugboard(pairs);
    Reflector reflector = B;

    std::vector<Setting> results;

    for (int i = 0; i < 26; i++)
    {
        for (int j = 0; j < 26; j++)
        {
            for (int k = 0; k < 26; k++)
            {
                Setting setting{{firstRotor, secondRotor, thirdRotor}, {i, j, k}};

                try_enigma(setting, cipher);

                results.push_back(setting);
            }
        }
    }

    return results;
}

bool operator<(const Setting &left, const Setting &right)
{
    return left.score < right.score;
}

void adjust_ring_settings(std::vector<Setting> &previousSettings, const std::string &cipher);
void adjust_plugboard(std::vector<Setting> &previousSettings, const std::string &cipher);
void finalAdjust(Setting &previousSetting, const std::string &cipher);
void find_all_solutions(Setting &previousSetting, const std::string &cipher);

void find_rotor_order(const std::string &cipher, int first_disk)
{
    std::vector<std::future<std::vector<Setting>>> asyncResults;
    std::vector<Setting> results;
    for (int i = 0; i < 8; i++)
    {
        // int i = first_disk;
        for (int j = 0; j < 8; j++)
        {
            if (i == j)
                continue;
            for (int k = 0; k < 8; k++)
            {
                if (k == i || k == j)
                    continue;
                if (i < 5 && j < 5 && k < 5)
                    continue;
                std::cout << "Trying " << rotorNames[i] << ", " << rotorNames[j] << ", " << rotorNames[k] << std::endl;

                asyncResults.push_back(std::async(try_all_keys, i, j, k, cipher));
            }
        }
        for (auto &&result : asyncResults)
        {
            result.wait();
            std::vector<Setting> resultVector = result.get();
            results.insert(results.end(), resultVector.begin(), resultVector.end());
        }
        asyncResults.clear();
    }

    std::sort(results.begin(), results.end(), [](const auto &left, const auto &right)
              { return right < left; });

    for (int i = 0; i < std::min(results.size(), (size_t)20); i++)
    {
        std::cout << i + 1 << " : " << results[i] << std::endl;
    }

    adjust_ring_settings(results, cipher);
}

Setting try_adjust_ring(Setting initial, const std::string &cipher)
{
    for (int j = 0; j < 26; j++)
    {
        for (int k = 0; k < 26; k++)
        {
            for (int l = 0; l < 26; l++)
            {
                Setting settingCopy = initial;
                settingCopy.ringSettings = {j, k, l};
                try_enigma(settingCopy, cipher);
                if (settingCopy.score > initial.score)
                {
                    initial = settingCopy;
                }
            }
        }
    }

    return initial;
}

void adjust_ring_settings(std::vector<Setting> &previousSettings, const std::string &cipher)
{
    std::cout << "ADJUSTING RING SETTINGS !!!" << std::endl;

    std::vector<std::future<Setting>> newSettingsFuture;
    std::vector<Setting> newSettings;

    for (int i = 0; i < std::min(previousSettings.size(), NUMBER_TO_HOLD); i++)
    {
        newSettingsFuture.push_back(std::async(try_adjust_ring, previousSettings[i], cipher));
    }

    for (auto &&result : newSettingsFuture)
    {
        result.wait();
        newSettings.push_back(result.get());
    }

    std::sort(newSettings.begin(), newSettings.end(), [](const auto &left, const auto &right)
              { return right < left; });

    for (int i = 0; i < std::min(newSettings.size(), (size_t)20); i++)
    {
        std::cout << i + 1 << " : " << newSettings[i] << std::endl;
    }

    adjust_plugboard(newSettings, cipher);
}

Setting try_adjust_plugboard(Setting initial, const std::string &cipher)
{
    std::vector<std::string> pairs;
    std::array<char, 10> usedLetters;
    usedLetters.fill('\0');

    try_enigma(initial, cipher, true);

    size_t pairIndex = 0;
    for (size_t i = 0; i < 26; i++)
    {
        if (pairIndex > 9)
            break;
        for (size_t j = 0; j < 26; j++)
        {
            if (i >= j)
                continue;
            if (std::find(usedLetters.begin(), usedLetters.end(), j + 'A') != usedLetters.end())
                continue;

            if (pairIndex > 9)
                break;
            if (pairIndex + 1 > pairs.size())
            {
                std::string str;
                str += i + 'A';
                str += i + 'A';
                pairs.push_back(str);
            }

            Setting copy = initial;
            std::vector<std::string> pairsCopy = pairs;

            pairsCopy[pairIndex][0] = i + 'A';
            pairsCopy[pairIndex][1] = j + 'A';

            copy.plugs = pairsCopy;
            try_enigma(copy, cipher, true);
            if (copy.score > initial.score)
            {
                initial = copy;
                pairs = pairsCopy;
                usedLetters[pairIndex] = pairs[pairIndex][1];
            }
        }
        if (pairIndex + 1 <= pairs.size() && pairs[pairIndex][0] != pairs[pairIndex][1])
            pairIndex++;
    }

    return initial;
}

void adjust_plugboard(std::vector<Setting> &previousSettings, const std::string &cipher)
{
    std::cout << "ADJUSTING PLUGBOARD !!!" << std::endl;

    std::vector<std::future<Setting>> newSettingsFuture;
    std::vector<Setting> newSettings;

    for (int i = 0; i < std::min(previousSettings.size(), NUMBER_TO_HOLD); i++)
    {
        newSettingsFuture.push_back(std::async(try_adjust_plugboard, previousSettings[i], cipher));
        // newSettings.push_back(try_adjust_plugboard(previousSettings[i], cipher));
    }

    for (auto &&result : newSettingsFuture)
    {
        result.wait();
        newSettings.push_back(result.get());
    }

    std::sort(newSettings.begin(), newSettings.end(), [](const auto &left, const auto &right)
              { return right < left; });

    for (int i = 0; i < std::min(newSettings.size(), (size_t)20); i++)
    {
        std::cout << i + 1 << " : " << newSettings[i] << std::endl;
    }

    finalAdjust(newSettings[0], cipher);
}

Setting try_final(Setting initial, const std::string &cipher)
{
    for (int k = 0; k < 26; k++)
    {
        for (int l = 0; l < 26; l++)
        {
            Setting copy = initial;
            copy.rotorPositions[1] = k;
            copy.rotorPositions[2] = l;

            try_enigma(copy, cipher, true);

            if (copy.score > initial.score)
            {
                initial = copy;
            }
        }
        // std::cout << "Trying with k = " << k << std::endl;
    }

    return initial;
}

void finalAdjust(Setting &previousSetting, const std::string &cipher)
{
    std::cout << "FINAL ADJUST!!!" << std::endl;
    std::vector<std::future<Setting>> futureResults;
    std::vector<Setting> results;

    try_enigma(previousSetting, cipher, true);
    for (int i = 0; i < 26; i++)
    {
        for (int j = 0; j < 26; j++)
        {
            Setting newSetting = previousSetting;
            newSetting.ringSettings[1] = i;
            newSetting.ringSettings[2] = j;

            futureResults.push_back(std::async(try_final, newSetting, cipher));
        }
    }

    for (auto &&result : futureResults)
    {
        result.wait();
        results.push_back(result.get());
    }

    std::sort(results.begin(), results.end(), [](const auto &left, const auto &right)
              { return right < left; });

    for (int i = 0; i < std::min(results.size(), (size_t)20); i++)
    {
        std::cout << i + 1 << " : " << results[i] << std::endl;
    }

    find_all_solutions(results[0], cipher);
}

void find_all_solutions(Setting &previousSetting, const std::string &cipher)
{
    std::ofstream results("results.txt");

    for (int i = 0; i < 26; i++)
    {
        previousSetting.ringSettings[0] = i;
        previousSetting.score = MINFINITY;
        for (int j = 0; j < 26; j++)
        {
            Setting copy = previousSetting;
            copy.rotorPositions[0] = j;
            try_enigma(copy, cipher, true);
            if (copy.score > previousSetting.score)
            {
                previousSetting = copy;
            }
        }
        previousSetting.print(results, false);
        results << std::endl;
    }
}

Setting strToSetting(const std::string &str)
{
    std::map<std::string, int> rotorMap = {
        {"I", 0},
        {"II", 1},
        {"III", 2},
        {"IV", 3},
        {"V", 4},
        {"VI", 5},
        {"VII", 6},
        {"VIII", 7}};

    Setting setting;

    std::stringstream sstream(str);

    std::string left_rotor_n;
    std::string center_rotor_n;
    std::string right_rotor_n;

    std::getline(sstream, left_rotor_n, ' ');
    std::getline(sstream, center_rotor_n, ' ');
    std::getline(sstream, right_rotor_n, ' ');

    setting.rotorIndexes = {rotorMap.at(left_rotor_n), rotorMap.at(center_rotor_n), rotorMap.at(right_rotor_n)};

    std::string tmp;
    std::getline(sstream, tmp, ' ');

    setting.ringSettings = {tmp[0] - 'A', tmp[1] - 'A', tmp[2] - 'A'};

    for (int i = 0; i < 10; i++)
    {
        std::getline(sstream, tmp, ' ');
        setting.plugs.push_back(tmp);
    }

    std::getline(sstream, tmp, ' ');

    setting.rotorPositions = {tmp[0] - 'A', tmp[1] - 'A', tmp[2] - 'A'};

    std::cout << "Setting : " << setting << std::endl;
    return setting;
}

int main(int argc, char **argv)
{
    std::stringstream sstream;
    sstream << std::cin.rdbuf();
    std::string cipher = sstream.str();

    cleanCipher(cipher);

    if (argc < 2)
    {
        std::cout << "You must provide first motor index";
        return -1;
    }

    find_rotor_order(cipher, std::atoi(argv[1]));
    return 0;

    // std::array<int, 3> rotorIndexes{0, 0, 0};
    // std::array<int, 3> rotorPositions{0, 0, 0};
    // std::array<int, 3> ringSettings{0, 0, 0};

    std::string setting_string = "VII IV I ONH AH BO CL DU FW GM JY KZ PV QX PDL";

    Setting setting = strToSetting(setting_string);

    // std::vector<Setting> settings{setting};

    finalAdjust(setting, cipher);
}
